##
#
module IdentifiersValid
  ISIN_FORMAT = '^[A-Z]{2}[A-Z0-9]{9}[0-9]$'.freeze
  SEDOL_FORMAT = '^[A-Z0-9]{6}\d{1}$'.freeze
  CUSIP_FORMAT = '^[A-Z0-9]{8}\d{1}$'.freeze
  ISIN_REGEX = Regexp.new(ISIN_FORMAT).freeze
  LETTER_TO_CODE_VALUE = 55 # Ascii code for 'A' == 65 && 'A' == 10 in isin calculation code
  CUSIP_SPECIAL_CHAR = { '*' => 36, '@' => 37, "\#" => 38 }.freeze
  SEDOL_VALID_CHAR = '0123456789BCDFGHJKLMNPQRSTVWXYZ'.freeze
  SEDOL_WEIGHT = [1, 3, 1, 7, 3, 9].freeze

  def self.identifier_valid?(value, field_key)
    case field_key.to_s.upcase
    when 'ISIN'
      isin_valid?(value)
    when 'SEDOL'
      sedol_valid?(value)
    when 'CUSIP'
      cusip_valid?(value)
    end
  end

  def self.identifier_invalid?(value, field_key)
    !identifier_valid?(value, field_key)
  end

  def self.cusip_valid?(cusip)
    return nil if cusip.nil? || cusip.size != 9

    proc_cusip = cusip.split(//)
    key = proc_cusip.pop
    sum = proc_cusip.each_with_index.map do |char, i|
      v = check_cusip_char(char, i)
      (v / 10).to_i + (v % 10)
    end.inject(:+)

    valid?(sum, key) ? cusip : nil
  end

  def self.check_cusip_char(char, index)
    v = if char.match?(/\d/)
          char.to_i
        elsif CUSIP_SPECIAL_CHAR[char]
          CUSIP_SPECIAL_CHAR[char]
        else
          char.ord - LETTER_TO_CODE_VALUE + 9
        end
    (index % 2).zero? ? v : v * 2
  end

  def self.cusip_invalid?(cusip)
    !cusip_valid?(cusip)
  end

  def self.sedol_valid?(sedol)
    return nil if sedol.nil? || sedol.size != 7

    proc_sedol = sedol.split(//)
    proc_sedol.each { |char| return nil unless SEDOL_VALID_CHAR.include?(char) }
    key = proc_sedol.pop
    sum = proc_sedol.zip(SEDOL_WEIGHT)
                    .map { |ch, weight| ch.to_i(36) * weight }.inject(:+)
    valid?(sum, key) ? sedol : nil
  end

  def self.sedol_invalid?(sedol)
    !sedol_valid?(sedol)
  end

  def self.isin_valid?(isin)
    return nil if isin.nil? || !isin.match?(ISIN_REGEX)

    proc_isin = isin.split(//)
    key = proc_isin.pop.to_i
    proc_isin = proc_isin.map do |e|
      e.match?(/[A-Z]/) ? (e.ord - LETTER_TO_CODE_VALUE).to_s.split(//) : e
    end.flatten.reverse
    proc_isin.each_index do |i|
      proc_isin[i] = (proc_isin[i].to_i * 2).to_s.split(//) if (i % 2).zero?
    end.flatten!
    ((key + proc_isin.map(&:to_i).inject(:+)) % 10).zero? ? isin : nil
  end

  def self.isin_invalid?(isin)
    !isin_valid?(isin)
  end

  def self.valid?(sum, key)
    ((10 - (sum % 10)) % 10).to_s == key
  end
end
